==> upload-shared-iso-1.24.0.tgz: package/src/statistics/storage/StatisticsCalculator.ts <==
import { groupBy, map, mapObjIndexed, pipe, sum } from "ramda";
import { StrictOmit } from "ts-essentials";
import { UploadStatisticsMimeRecord } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/UploadStatisticsMimeRecord";
import { StorageStatisticsSegmentDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/StorageStatisticsSegmentDto";
import { UploadStatisticsDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/UploadStatisticsDto";
import { UploadStatisticsMimeSegment } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/UploadStatisticsMimeSegment";
import { TypedFileEvent } from "@upload-io/upload-shared-iso/statistics/storage/TypedFileEvent";
import {
  BigSizeInBytes,
  EpochMillis,
  Milliseconds,
  SizeInBytes
} from "@upload-io/upload-shared-iso/common/model/Primitives";
import { NonEmpty, NonEmptyArray } from "@upload-io/upload-shared-iso/common/model/NonEmptyArray";
import { BasicFileEvent } from "@upload-io/upload-shared-iso/statistics/storage/BasicFileEvent";
import { IsoTimeUtils } from "@upload-io/upload-shared-iso/common/IsoTimeUtils";
import { BoxedArithmetic } from "@upload-io/upload-shared-iso/common/BoxedArithmetic";
import { BytesByEpochDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/BytesByEpochDto";
import { DailyMonthlyBytesDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/DailyMonthlyBytesDto";
import { StorageStatisticsByTimeDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/StorageStatisticsByTimeDto";
import { InvocationEvent } from "@upload-io/upload-shared-iso/statistics/storage/InvocationEvent";
import { DurationByEpochDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/DurationByEpochDto";
import { DailyMonthlyDurationsDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/DailyMonthlyDurationsDto";

const daysToRecord = 60; // UI requires at least 30 for 'last 30 days' widget. We capture more for good measure.
const monthsToRecord = 24; // UI requires at least 12 for 'monthly bandwidth' widget. We capture more for good measure.

type Predicate<T> = (item: T) => boolean;
type MimeFilters = StrictOmit<UploadStatisticsMimeRecord<Predicate<TypedFileEvent>>, "other">;
type EventsByMime = UploadStatisticsMimeRecord<TypedFileEvent[]>;
type StatisticsByMime = UploadStatisticsMimeRecord<StorageStatisticsSegmentDto>;
type EventsFunc<T> = (events: TypedFileEvent[]) => T;

const mimeFilters: MimeFilters = {
  audio: x => x.mime.startsWith("audio/"),
  image: x => x.mime.startsWith("image/"),
  video: x => x.mime.startsWith("video/"),
  document: x =>
    [
      "application/x-abiword",
      "application/msword",
      "application/vnd.ms-",
      "application/vnd.openxmlformats",
      "application/vnd.oasis.opendocument",
      "application/pdf"
    ].some(prefix => x.mime.startsWith(prefix))
};

const mimeFiltersEntries: Array<[UploadStatisticsMimeSegment, Predicate<TypedFileEvent>]> = Object.entries(
  mimeFilters
).map(([k, v]) => [k as UploadStatisticsMimeSegment, v]);

export class StatisticsCalculator {
  static calculateStatistics(newEvents: NonEmptyArray<TypedFileEvent>, valueMultiplier: 1 | -1): UploadStatisticsDto {
    const groupByMime: EventsFunc<EventsByMime> = groupBy(event => {
      const match = mimeFiltersEntries.find(([_, filter]) => filter(event));
      if (match !== undefined) {
        return match[0];
      }
      return "other";
    });

    const calculateStatisticsForMime = map<EventsByMime, StatisticsByMime>(
      (events: TypedFileEvent[]) => this.calculateStatisticsSegment(NonEmpty(events), valueMultiplier) // Input is from a groupBy, so we know is non-empty.
    );

    const newStatisticsMime: Partial<StatisticsByMime> = pipe(groupByMime, calculateStatisticsForMime)(newEvents);

    return {
      all: this.calculateStatisticsSegment(newEvents, valueMultiplier),
      byMime: {
        ...UploadStatisticsDto.empty.byMime,
        ...newStatisticsMime
      },
      byTime: this.calculateStorageStatisticsByTime(newEvents, valueMultiplier)
    };
  }

  static calculateMonthlyDailyBytes(events: BasicFileEvent[]): DailyMonthlyBytesDto {
    const now = EpochMillis(Date.now());
    return {
      byDay: this.calculateBytesPerEpochUnit(IsoTimeUtils.toEpochDay, daysToRecord, events, now),
      byMonth: this.calculateBytesPerEpochUnit(IsoTimeUtils.toEpochMonth, monthsToRecord, events, now)
    };
  }

  static calculateMonthlyDailyDurations(events: InvocationEvent[]): DailyMonthlyDurationsDto {
    const now = EpochMillis(Date.now());
    return {
      byDay: this.calculateDurationsPerEpochUnit(IsoTimeUtils.toEpochDay, daysToRecord, events, now),
      byMonth: this.calculateDurationsPerEpochUnit(IsoTimeUtils.toEpochMonth, monthsToRecord, events, now)
    };
  }

  static calculateStatisticsSegment(
    newEvents: NonEmptyArray<BasicFileEvent>,
    valueMultiplier: 1 | -1
  ): StorageStatisticsSegmentDto {
    const timestamps = newEvents.map(x => x.timeOfEvent);
    return {
      firstFileCreated: valueMultiplier === -1 ? undefined : EpochMillis(Math.min(...timestamps)),
      lastFileCreated: valueMultiplier === -1 ? undefined : EpochMillis(Math.max(...timestamps)),
      totalFileCount: valueMultiplier * newEvents.length,
      totalFileSize: SizeInBytes(valueMultiplier * sum(newEvents.map(x => x.size))) // Can go negative (in a shard)
    };
  }

  static calculateStorageStatisticsByTime(
    newEvents: NonEmptyArray<BasicFileEvent>,
    valueMultiplier: 1 | -1
  ): StorageStatisticsByTimeDto {
    return {
      creates: valueMultiplier === 1 ? this.calculateMonthlyDailyBytes(newEvents) : DailyMonthlyBytesDto.empty,
      deletes: valueMultiplier === -1 ? this.calculateMonthlyDailyBytes(newEvents) : DailyMonthlyBytesDto.empty
    };
  }

  private static calculateBytesPerEpochUnit(
    getEpochUnit: (time: EpochMillis) => number,
    keepLastNUnits: number,
    events: BasicFileEvent[],
    now: EpochMillis
  ): BytesByEpochDto {
    const eventsByEpoch = groupBy<BasicFileEvent>(x => getEpochUnit(x.timeOfEvent).toString(), events);
    const totalSize = (events: BasicFileEvent[]): BigSizeInBytes =>
      BoxedArithmetic.bigSizeInBytes.sum(events.map(event => BigSizeInBytes(event.size.toString())));

    return IsoTimeUtils.trimEpochKeys(getEpochUnit, now, keepLastNUnits, mapObjIndexed(totalSize, eventsByEpoch));
  }

  private static calculateDurationsPerEpochUnit(
    getEpochUnit: (time: EpochMillis) => number,
    keepLastNUnits: number,
    events: InvocationEvent[],
    now: EpochMillis
  ): DurationByEpochDto {
    const eventsByEpoch = groupBy<InvocationEvent>(x => getEpochUnit(x.timeOfEvent).toString(), events);
    const totalSize = (events: InvocationEvent[]): Milliseconds =>
      Milliseconds(sum(events.map(event => event.duration)));

    return IsoTimeUtils.trimEpochKeys(getEpochUnit, now, keepLastNUnits, mapObjIndexed(totalSize, eventsByEpoch));
  }
}

==> upload-shared-iso-1.31.0.tgz: package/src/statistics/storage/StatisticsCalculator.ts <==
import { groupBy, map, mapObjIndexed, pipe, sum } from "ramda";
import { StrictOmit } from "ts-essentials";
import { UploadStatisticsMimeRecord } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/UploadStatisticsMimeRecord";
import { StorageStatisticsSegmentDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/StorageStatisticsSegmentDto";
import { UploadStatisticsDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/UploadStatisticsDto";
import { UploadStatisticsMimeSegment } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/UploadStatisticsMimeSegment";
import { TypedFileEvent } from "@upload-io/upload-shared-iso/statistics/storage/TypedFileEvent";
import { BigSizeInBytes, EpochMillis } from "@upload-io/upload-shared-iso/common/model/Primitives";
import { Milliseconds, SizeInBytes } from "@upload-io/upload-shared-iso-lite";
import { NonEmpty, NonEmptyArray } from "@upload-io/upload-shared-iso/common/model/NonEmptyArray";
import { BasicFileEvent } from "@upload-io/upload-shared-iso/statistics/storage/BasicFileEvent";
import { IsoTimeUtils } from "@upload-io/upload-shared-iso/common/IsoTimeUtils";
import { BoxedArithmetic } from "@upload-io/upload-shared-iso/common/BoxedArithmetic";
import { BytesByEpochDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/BytesByEpochDto";
import { DailyMonthlyBytesDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/DailyMonthlyBytesDto";
import { StorageStatisticsByTimeDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/StorageStatisticsByTimeDto";
import { InvocationEvent } from "@upload-io/upload-shared-iso/statistics/storage/InvocationEvent";
import { DurationByEpochDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/DurationByEpochDto";
import { DailyMonthlyDurationsDto } from "@upload-io/upload-shared-iso/statistics/storage/dtos/shared/DailyMonthlyDurationsDto";

const daysToRecord = 60; // UI requires at least 30 for 'last 30 days' widget. We capture more for good measure.
const monthsToRecord = 24; // UI requires at least 12 for 'monthly bandwidth' widget. We capture more for good measure.

type Predicate<T> = (item: T) => boolean;
type MimeFilters = StrictOmit<UploadStatisticsMimeRecord<Predicate<TypedFileEvent>>, "other">;
type EventsByMime = UploadStatisticsMimeRecord<TypedFileEvent[]>;
type StatisticsByMime = UploadStatisticsMimeRecord<StorageStatisticsSegmentDto>;
type EventsFunc<T> = (events: TypedFileEvent[]) => T;

const mimeFilters: MimeFilters = {
  audio: x => x.mime.startsWith("audio/"),
  image: x => x.mime.startsWith("image/"),
  video: x => x.mime.startsWith("video/"),
  document: x =>
    [
      "application/x-abiword",
      "application/msword",
      "application/vnd.ms-",
      "application/vnd.openxmlformats",
      "application/vnd.oasis.opendocument",
      "application/pdf"
    ].some(prefix => x.mime.startsWith(prefix))
};

const mimeFiltersEntries: Array<[UploadStatisticsMimeSegment, Predicate<TypedFileEvent>]> = Object.entries(
  mimeFilters
).map(([k, v]) => [k as UploadStatisticsMimeSegment, v]);

export class StatisticsCalculator {
  static calculateStatistics(newEvents: NonEmptyArray<TypedFileEvent>, valueMultiplier: 1 | -1): UploadStatisticsDto {
    const groupByMime: EventsFunc<EventsByMime> = groupBy(event => {
      const match = mimeFiltersEntries.find(([_, filter]) => filter(event));
      if (match !== undefined) {
        return match[0];
      }
      return "other";
    });

    const calculateStatisticsForMime = map<EventsByMime, StatisticsByMime>(
      (events: TypedFileEvent[]) => this.calculateStatisticsSegment(NonEmpty(events), valueMultiplier) // Input is from a groupBy, so we know is non-empty.
    );

    const newStatisticsMime: Partial<StatisticsByMime> = pipe(groupByMime, calculateStatisticsForMime)(newEvents);

    return {
      all: this.calculateStatisticsSegment(newEvents, valueMultiplier),
      byMime: {
        ...UploadStatisticsDto.empty.byMime,
        ...newStatisticsMime
      },
      byTime: this.calculateStorageStatisticsByTime(newEvents, valueMultiplier)
    };
  }

  static calculateMonthlyDailyBytes(events: BasicFileEvent[]): DailyMonthlyBytesDto {
    const now = EpochMillis(Date.now());
    return {
      byDay: this.calculateBytesPerEpochUnit(IsoTimeUtils.toEpochDay, daysToRecord, events, now),
      byMonth: this.calculateBytesPerEpochUnit(IsoTimeUtils.toEpochMonth, monthsToRecord, events, now)
    };
  }

  static calculateMonthlyDailyDurations(events: InvocationEvent[]): DailyMonthlyDurationsDto {
    const now = EpochMillis(Date.now());
    return {
      byDay: this.calculateDurationsPerEpochUnit(IsoTimeUtils.toEpochDay, daysToRecord, events, now),
      byMonth: this.calculateDurationsPerEpochUnit(IsoTimeUtils.toEpochMonth, monthsToRecord, events, now)
    };
  }

  static calculateStatisticsSegment(
    newEvents: NonEmptyArray<BasicFileEvent>,
    valueMultiplier: 1 | -1
  ): StorageStatisticsSegmentDto {
    const timestamps = newEvents.map(x => x.timeOfEvent);
    return {
      firstFileCreated: valueMultiplier === -1 ? undefined : EpochMillis(Math.min(...timestamps)),
      lastFileCreated: valueMultiplier === -1 ? undefined : EpochMillis(Math.max(...timestamps)),
      totalFileCount: valueMultiplier * newEvents.length,
      totalFileSize: SizeInBytes(valueMultiplier * sum(newEvents.map(x => x.size))) // Can go negative (in a shard)
    };
  }

  static calculateStorageStatisticsByTime(
    newEvents: NonEmptyArray<BasicFileEvent>,
    valueMultiplier: 1 | -1
  ): StorageStatisticsByTimeDto {
    return {
      creates: valueMultiplier === 1 ? this.calculateMonthlyDailyBytes(newEvents) : DailyMonthlyBytesDto.empty,
      deletes: valueMultiplier === -1 ? this.calculateMonthlyDailyBytes(newEvents) : DailyMonthlyBytesDto.empty
    };
  }

  private static calculateBytesPerEpochUnit(
    getEpochUnit: (time: EpochMillis) => number,
    keepLastNUnits: number,
    events: BasicFileEvent[],
    now: EpochMillis
  ): BytesByEpochDto {
    const eventsByEpoch = groupBy<BasicFileEvent>(x => getEpochUnit(x.timeOfEvent).toString(), events);
    const totalSize = (events: BasicFileEvent[]): BigSizeInBytes =>
      BoxedArithmetic.bigSizeInBytes.sum(events.map(event => BigSizeInBytes(event.size.toString())));

    return IsoTimeUtils.trimEpochKeys(getEpochUnit, now, keepLastNUnits, mapObjIndexed(totalSize, eventsByEpoch));
  }

  private static calculateDurationsPerEpochUnit(
    getEpochUnit: (time: EpochMillis) => number,
    keepLastNUnits: number,
    events: InvocationEvent[],
    now: EpochMillis
  ): DurationByEpochDto {
    const eventsByEpoch = groupBy<InvocationEvent>(x => getEpochUnit(x.timeOfEvent).toString(), events);
    const totalSize = (events: InvocationEvent[]): Milliseconds =>
      Milliseconds(sum(events.map(event => event.duration)));

    return IsoTimeUtils.trimEpochKeys(getEpochUnit, now, keepLastNUnits, mapObjIndexed(totalSize, eventsByEpoch));
  }
}
